/*
 * PS3 Media Server, for streaming any medias to your PS3.
 * Copyright (C) 2008  A.Brochard
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License only.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package net.pms.encoders;

import java.awt.Font;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;

import javax.swing.JCheckBox;
import javax.swing.JComponent;

import net.pms.PMS;
import net.pms.configuration.PmsConfiguration;
import net.pms.configuration.RendererConfiguration;
import net.pms.dlna.DLNAMediaAudio;
import net.pms.dlna.DLNAMediaInfo;
import net.pms.dlna.InputFile;
import net.pms.formats.Format;
import net.pms.io.OutputParams;
import net.pms.io.PipeIPCProcess;
import net.pms.io.PipeProcess;
import net.pms.io.ProcessWrapper;
import net.pms.io.ProcessWrapperImpl;
import net.pms.io.StreamModifier;
import net.pms.util.Iso639;

import com.jgoodies.forms.builder.PanelBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.CellConstraints;
import com.jgoodies.forms.layout.FormLayout;

import com.sun.jna.Platform;

public class AvcmodVideo extends Player {

	public static final String ID = "avcmod"; //$NON-NLS-1$
	private PmsConfiguration configuration;
	
	public AvcmodVideo(PmsConfiguration configuration) {
		this.configuration = configuration;
	}
	
	public boolean excludeFormat(Format extension) {
		String m = extension.getMatchedId();
		return m != null && !m.equals("mkv") && !m.equals("ts") && !m.equals("tp") && !m.equals("m2ts") && !m.equals("m2t") && !m.equals("mpg") && !m.equals("evo") && !m.equals("mpeg")
			&& !m.equals("vob") && !m.equals("m2v") && !m.equals("mts");
	}
	
	@Override
	public int purpose() {
		return VIDEO_SIMPLEFILE_PLAYER;
	}
	
	@Override
	public String id() {
		return ID;
	}
	
	@Override
	public boolean isTimeSeekable() {
		return true;
	}

	@Override
	public String[] args() {
		return null;
	}

	@Override
	public String executable() {
		return configuration.getTsmuxerPath();
	}

	@Override
	public ProcessWrapper launchTranscode(String fileName, DLNAMediaInfo media, OutputParams params)
			throws IOException {
		if (media == null) {
			PMS.minimal("[avcmod] media is null! unable to transcode, exit.");
			return null;
		}
		setAudioAndSubs(fileName, media, params, configuration);
		
		PipeIPCProcess ffVideoPipe = null;
		ProcessWrapperImpl ffVideo = null;
		PipeIPCProcess[] ffAudioPipe = null;
		ProcessWrapperImpl[] ffAudio = null;
		String[] ffmpegLPCMextract = null;
		OutputParams ffparams = null;
		
		//params.waitbeforestart = 5000;
		params.manageFastStart();
		
		if (useFFmpegVideo) {
			ffVideoPipe = new PipeIPCProcess(System.currentTimeMillis() + "ffmpegvideo.h264",
				System.currentTimeMillis() + "videoout.h264", false, true);
			ffmpegLPCMextract = new String[] {
				configuration.getFfmpegPath(), "-i", fileName,
				//"-ss", params.timeseek > 0 ? String.valueOf(params.timeseek) : "0",
				"-vcodec", "copy", "-an",
				"-y", // overwrite output file for reconnect.
				ffVideoPipe.getInputPipe()
			};
			InputFile newInput = new InputFile();
			newInput.filename = fileName;
			newInput.push = params.stdin;
			
			media.isVideoPS3Compatible(newInput); // get h264_annexB
			if (media.h264_annexB != null && media.h264_annexB.length > 0) {
				StreamModifier sm = new StreamModifier();
				sm.setHeader(media.h264_annexB);
				sm.setH264_annexb(true);
				ffVideoPipe.setModifier(sm);
			} else {
				PMS.info("[avcmod] set h264_annexB header failed.");
			}
			ffparams = new OutputParams(configuration);
			ffparams.maxBufferSize = 1;
			ffparams.stdin = params.stdin;
			ffVideo = new ProcessWrapperImpl(ffmpegLPCMextract, ffparams);
		}
		
		if (media.audioCodes != null && media.audioCodes.size() > 0) {
			int numAudioTracks = 1;
			if (configuration.isMuxAllAudioTracks())
				numAudioTracks = media.audioCodes.size();
			ffAudioPipe = new PipeIPCProcess[numAudioTracks];
			ffAudio = new ProcessWrapperImpl[numAudioTracks];
			ffparams = new OutputParams(configuration);
			ffparams.maxBufferSize = 1;
			ffparams.stdin = params.stdin;
			for (int i = 0; i < numAudioTracks; i++) {
				DLNAMediaAudio audio = media.audioCodes.get(i);
				String ext = audio.isAC3() ? ".ac3" : ".aac";
				ffAudioPipe[i] = new PipeIPCProcess(System.currentTimeMillis() + "ffmpegaudio" + (1+i) + ext,
					System.currentTimeMillis() + "audioout" + (1+i) + ext, false, true);
				ffmpegLPCMextract = new String[] {
						configuration.getFfmpegPath(), "-i", fileName,
						"", "", //3,4: channel mapping (set later).
						"-acodec", "copy", "-vn", "-y", ffAudioPipe[i].getInputPipe()
					};
				if (i > 0) { // set -map option. first track is auto map.
					ffmpegLPCMextract[3] = "-map";
					// TODO should be container track id
					ffmpegLPCMextract[4] = "0." + (i + 1);
				}
				ffAudio[i] = new ProcessWrapperImpl(ffmpegLPCMextract, ffparams);
			}
		} else {
			PMS.info("[avcmod] no audio track found.");
		}
		
		File f = new File(configuration.getTempFolder(), "pms-tsmuxer.meta");
		params.log = false;
		PrintWriter pw = new PrintWriter(f, Platform.isWindows() ? "MS932" : "UTF-8");
		pw.print("MUXOPT --no-pcr-on-video-pid --new-audio-pes --no-asyncio");
		if (params.timeseek > 0) {
			pw.print(" --cut-start=" + params.timeseek + "s");
		}
		pw.println(" --vbr --vbv-len=500"); //$NON-NLS-1$
		String video_source;
		if (ffVideoPipe != null) {
			video_source = ffVideoPipe.getOutputPipe();
		} else {
			video_source = fileName;
		}
		String fps = "";
		String valid_fps = media.getValidFps(false);
		if (valid_fps != null)
			fps = "fps=" + valid_fps + ", ";
		String level = "";
		if (params.mediaRenderer != null && params.mediaRenderer.isH264Level41Limited())
			level = "level=4.1, ";
		pw.println("V_MPEG4/ISO/AVC, \"" + video_source + "\", " + fps + level + "insertSEI, contSPS, " + "track=1");
		if (ffAudioPipe != null) {
			for (int i = 0; i < ffAudioPipe.length; i++) {
				DLNAMediaAudio audio = media.audioCodes.get(i);
				String timeshift = "";
				if (audio != null && audio.delay != 0)
					timeshift = "timeshift=" + audio.delay + "ms, ";
				String lang = "";
				String valid_lang = Iso639.getISO639_2Code(audio.lang);
				if (valid_lang != null)
					lang = "lang=" + valid_lang + ", ";
				pw.println("A_" + audio.getAudioCodec() + ", \"" + ffAudioPipe[i].getOutputPipe() + "\", " + timeshift + lang + "track=" + (2+i));
			}
		}
		pw.close();
		
		OutputParams tsparams = new OutputParams(configuration);
		tsparams.maxBufferSize = 100;
		
		PipeProcess tsPipe = new PipeProcess(System.currentTimeMillis() + "tsmuxerout.ts", tsparams);
		String cmd [] = new String [] { executable(), f.getAbsolutePath(), tsPipe.getInputPipe() };
		ProcessWrapperImpl p = new ProcessWrapperImpl(cmd, params);
		params.maxBufferSize = 100;
		params.input_pipes[0] = tsPipe;
		params.stdin = null;
		ProcessWrapper pipe_process = tsPipe.getPipeProcess();
		p.attachProcess(pipe_process);
		pipe_process.runInNewThread();
		try {
			Thread.sleep(50);
		} catch (InterruptedException e) { }
		tsPipe.deleteLater();
		
		if (ffVideoPipe != null) {
			ProcessWrapper ff_pipe_process = ffVideoPipe.getPipeProcess();
			ffVideo.attachProcess(ff_pipe_process);
			ff_pipe_process.runInNewThread();
			try {
				Thread.sleep(50);
			} catch (InterruptedException e) { }
			ffVideoPipe.deleteLater();
			
			p.attachProcess(ffVideo);
			ffVideo.runInNewThread();
		}
		
		if (ffAudioPipe != null) {
			for(int i=0;i<ffAudioPipe.length;i++) {
				ProcessWrapper ff_pipe_process = ffAudioPipe[i].getPipeProcess();
				ffAudio[i].attachProcess(ff_pipe_process);
				ff_pipe_process.runInNewThread();
				try {
					Thread.sleep(50);
				} catch (InterruptedException e) { }
				ffAudioPipe[i].deleteLater();
				
				p.attachProcess(ffAudio[i]);
				ffAudio[i].runInNewThread();
			}
		}
		
		try {
			Thread.sleep(100);
		} catch (InterruptedException e) { }
		
		p.runInNewThread();
		return p;
	}

	@Override
	public String mimeType() {
		return "video/mpeg"; //$NON-NLS-1$
	}

	@Override
	public String name() {
		return "AvcmodVideo"; //$NON-NLS-1$
	}

	@Override
	public int type() {
		return Format.VIDEO;
	}
	
	private static boolean useFFmpegVideo = false;
	private JCheckBox useffmpegvideo;
	
	@Override
	public JComponent config() {
		FormLayout layout = new FormLayout(
				"left:pref, 0:grow", //$NON-NLS-1$
				"p, 3dlu, p, 3dlu, p, 3dlu, 0:grow"); //$NON-NLS-1$
		PanelBuilder builder = new PanelBuilder(layout);
		builder.setBorder(Borders.EMPTY_BORDER);
		builder.setOpaque(false);
		
		CellConstraints cc = new CellConstraints();
		JComponent cmp = builder.addSeparator(
				"AvcmodVideo Settings",  cc.xyw(2, 1, 1)); //$NON-NLS-1$
		cmp = (JComponent) cmp.getComponent(0);
		cmp.setFont(cmp.getFont().deriveFont(Font.BOLD));
		
		useffmpegvideo = new JCheckBox(
				"Use FFmpeg for extracting a video track."); //$NON-NLS-1$
		useffmpegvideo.setContentAreaFilled(false);
		useffmpegvideo.setSelected(useFFmpegVideo);
		useffmpegvideo.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				useFFmpegVideo = e.getStateChange() == ItemEvent.SELECTED;
			}
		});
		builder.add(useffmpegvideo, cc.xy(2, 3));
		
		return builder.getPanel();
	}
	
	public boolean isInternalSubtitlesSupported() {
		return false;
	}
	public boolean isExternalSubtitlesSupported() {
		return false;
	}

	@Override
	public boolean isPlayerCompatible(RendererConfiguration mediaRenderer) {
		return mediaRenderer != null && mediaRenderer.isMuxH264MpegTS();
	}
}
